(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Parameter identifier location : Set.

Inductive Constante : Set :=
  | Const_int : nat -> Constante
  | Const_bool : bool -> Constante
  | unit : Constante. 	

Inductive expr : Set :=
  | Const : Constante -> expr
  | Var : identifier -> expr
  | Loc : location -> expr
  | Fun : identifier -> expr -> expr
  | Apply : expr -> expr -> expr
  | Let_in : identifier -> expr -> expr -> expr
  | If : expr -> expr -> expr -> expr
  | While : expr -> expr -> expr
  | Ref : expr -> expr
  | Deref : expr -> expr
  | Sequ : expr -> expr -> expr
  | Assign : expr -> expr -> expr.

Definition vrai := Const (Const_bool true).
Definition faux := Const (Const_bool false).


Inductive free_ident : identifier -> expr -> Prop :=
  | free_var : forall x : identifier, free_ident x (Var x)
  | free_apply1 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e1 -> free_ident x (Apply e1 e2)
  | free_apply2 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e2 -> free_ident x (Apply e1 e2)
  | free_let1 :
      forall (x y : identifier) (e1 e2 : expr),
      free_ident x e1 -> free_ident x (Let_in y e1 e2)
  | free_let2 :
      forall (x y : identifier) (e1 e2 : expr),
      free_ident x e2 -> x <> y -> free_ident x (Let_in y e1 e2)
  | free_fun :
      forall (x y : identifier) (e1 : expr),
      free_ident x e1 -> x <> y -> free_ident x (Fun y e1)
  | free_if1 :
      forall (x : identifier) (e e1 e2 : expr),
      free_ident x e -> free_ident x (If e e1 e2)
  | free_if2 :
      forall (x : identifier) (e e1 e2 : expr),
      free_ident x e1 -> free_ident x (If e e1 e2)
  | free_if3 :
      forall (x : identifier) (e e1 e2 : expr),
      free_ident x e2 -> free_ident x (If e e1 e2)
  | free_while1 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e1 -> free_ident x (While e1 e2)
  | free_while2 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e2 -> free_ident x (While e1 e2)
  | free_ref :
      forall (x : identifier) (e : expr),
      free_ident x e -> free_ident x (Ref e)
  | free_deref :
      forall (x : identifier) (e : expr),
      free_ident x e -> free_ident x (Deref e)
  | free_sequ1 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e1 -> free_ident x (Sequ e1 e2)
  | free_sequ2 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e2 -> free_ident x (Sequ e1 e2)
  | free_assign1 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e1 -> free_ident x (Assign e1 e2)
  | free_assign2 :
      forall (x : identifier) (e1 e2 : expr),
      free_ident x e2 -> free_ident x (Assign e1 e2).


(* (subst_expr e x e' e'')  means e[e'\x]=e'' *)
Inductive subst_expr : expr -> identifier -> expr -> expr -> Prop :=
  | subst_const :
      forall (x : identifier) (e' : expr) (c : Constante),
      subst_expr (Const c) x e' (Const c)
  | subst_var :
      forall (x : identifier) (e' : expr), subst_expr (Var x) x e' e'
  | subst_nvar :
      forall (x y : identifier) (e' : expr),
      x <> y -> subst_expr (Var y) x e' (Var y)
  | subst_app :
      forall (x : identifier) (e' e1 e2 e1' e2' : expr),
      subst_expr e1 x e' e1' ->
      subst_expr e2 x e' e2' -> subst_expr (Apply e1 e2) x e' (Apply e1' e2')
  | subst_fun :
      forall (x : identifier) (e e' : expr),
      subst_expr (Fun x e) x e' (Fun x e)
  | subst_nfun :
      forall (x y : identifier) (e e' e'' : expr),
      ~ free_ident y e' ->
      x <> y ->
      subst_expr e x e' e'' -> subst_expr (Fun y e) x e' (Fun y e'')
  | subst_let :
      forall (x : identifier) (e' p e e1 : expr),
      subst_expr e x e' e1 -> subst_expr (Let_in x e p) x e' (Let_in x e1 p)
  | subst_nlet :
      forall (x y : identifier) (e' p p'' e e'' : expr),
      ~ free_ident y e' ->
      x <> y ->
      subst_expr e x e' e'' ->
      subst_expr p x e' p'' ->
      subst_expr (Let_in y e p) x e' (Let_in y e'' p'')
  | subst_if :
      forall (x : identifier) (eb e1 e2 e' eb' e1' e2' : expr),
      subst_expr eb x e' eb' ->
      subst_expr e1 x e' e1' ->
      subst_expr e2 x e' e2' ->
      subst_expr (If eb e1 e2) x e' (If eb' e1' e2')
  | subst_while :
      forall (x : identifier) (e1 e2 e' e1' e2' : expr),
      subst_expr e1 x e' e1' ->
      subst_expr e2 x e' e2' -> subst_expr (While e1 e2) x e' (While e1' e2')
  | subst_ref :
      forall (x : identifier) (e e' er : expr),
      subst_expr e x e' er -> subst_expr (Ref e) x e' (Ref er)
  | subst_deref :
      forall (x : identifier) (e e' er : expr),
      subst_expr e x e' er -> subst_expr (Deref e) x e' (Deref er)
  | subst_sequ :
      forall (x : identifier) (e1 e2 er1 er2 e' : expr),
      subst_expr e1 x e' er1 ->
      subst_expr e2 x e' er2 -> subst_expr (Sequ e1 e2) x e' (Sequ er1 er2)
  | subst_assign :
      forall (x : identifier) (e1 e2 er1 er2 e' : expr),
      subst_expr e1 x e' er1 ->
      subst_expr e2 x e' er2 ->
      subst_expr (Assign e1 e2) x e' (Assign er1 er2)
  | subst_loc :
      forall (l : location) (x : identifier) (e' : expr),
      subst_expr (Loc l) x e' (Loc l).
