(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec.
Require Import Table.
Require Import Typing_spec.
Require Import Small_step_spec.
Require Import Subject_reduction_proof. (*pour memory_respects_env*)
Require Import Is_value_dec_proof.



Lemma free_ident_is_in_context :
 forall (x : identifier) (e : expr) (envx : env_typ_var) 
   (envl : env_typ_loc) (t : type),
 type_of (envx, envl) e t -> free_ident x e -> in_dom x envx.

intros x e.
induction e
 as
  [c|
   i|
   l|
   i e Hrece|
   e1 Hrece1 e0 Hrece0|
   i e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0 e2 Hrece2|
   e1 Hrece1 e0 Hrece0|
   e Hrece|
   e Hrece|
   e1 Hrece1 e0 Hrece0|
   e1 Hrece1 e0 Hrece0]; intros envx envl t Htyp Hfree; 
 inversion Hfree; inversion Htyp.

inversion H5; simpl in |- *; trivial.

apply add_dom with (a' := i) (b := t0).
assumption.

apply Hrece with (envl := envl) (t := t'); assumption.

apply Hrece1 with (envl := envl) (t := Arrow t0 t); assumption.

apply Hrece0 with (envl := envl) (t := t0); assumption.

apply Hrece1 with (envl := envl) (t := t0); assumption.

apply add_dom with (a' := i) (b := t0).
assumption.

apply Hrece0 with (envl := envl) (t := t); assumption.

apply Hrece1 with (envl := envl) (t := Const_t Bool); assumption.

apply Hrece0 with (envl := envl) (t := t); assumption.

apply Hrece2 with (envl := envl) (t := t); assumption.

apply Hrece1 with (envl := envl) (t := Const_t Bool); assumption.

apply Hrece0 with (envl := envl) (t := t0); assumption.

apply Hrece with (envl := envl) (t := t0); assumption.

apply Hrece with (envl := envl) (t := Ref_t t); assumption.

apply Hrece1 with (envl := envl) (t := t'); assumption.

apply Hrece0 with (envl := envl) (t := t); assumption.

apply Hrece1 with (envl := envl) (t := Ref_t t0); assumption.

apply Hrece0 with (envl := envl) (t := t0); assumption.

Qed.

(*------------------*)
Lemma type_of_empty_ground :
 forall (e : expr) (t : type) (envl : env_typ_loc) (f : identifier -> type),
 type_of (empty_table f, envl) e t -> forall x : identifier, ~ free_ident x e.

intros e t envl f Htyp x.
intro Hfree.
absurd (in_dom x (empty_table f)).
intro in_empty.
inversion in_empty.

apply free_ident_is_in_context with (e := e) (envl := envl) (t := t);
 assumption.
Qed.
(*------------------*)




Lemma substitute_succeeds :
 forall (e e1 : expr) (i : identifier) (t : type) (envl : env_typ_loc)
   (f : identifier -> type),
 type_of (empty_table f, envl) e1 t -> exists e' : expr, subst_expr e i e1 e'.

simple induction e; intros.
(*const*)
exists (Const c).
apply subst_const.

(*var*)
elim (eq_A_dec i i0).
intro Heq; rewrite Heq.
exists e1.
apply subst_var.

intro Hneq.
exists (Var i).
apply subst_nvar.
auto.

(*loc*)
exists (Loc l).
apply subst_loc.

(*fun*)
elim (eq_A_dec i i0); intro Heq.
rewrite Heq.
exists (Fun i0 e0).
apply subst_fun.

elim (H e1 i0 t envl f H0).
intros e0' Hdef_e0'.
exists (Fun i e0').
apply subst_nfun.
apply type_of_empty_ground with (envl := envl) (f := f) (t := t).
assumption.
 
auto.

assumption.

(*apply*)
elim (H e2 i t envl f H1).
intros e0' Hdef_e0'.
elim (H0 e2 i t envl f H1).
intros e1' Hdef_e1'.
exists (Apply e0' e1').
apply subst_app; assumption.

(*let_in*)
elim (H e2 i0 t envl f H1).
intros e0' Hdef_e0'.
elim (eq_A_dec i i0); intros Heq.
rewrite Heq.
exists (Let_in i0 e0' e1).
apply subst_let; assumption.

elim (H0 e2 i0 t envl f H1).
intros e1' Hdef_e1'.
exists (Let_in i e0' e1').
apply subst_nlet; auto.
apply type_of_empty_ground with (t := t) (f := f) (envl := envl); auto.

(*if*)
elim (H e3 i t envl f H2).
intros e0' Hdef_e0'.
elim (H1 e3 i t envl f H2).
intros e2' Hdef_e2'.
elim (H0 e3 i t envl f H2).
intros e1' Hdef_e1'.
exists (If e0' e1' e2').
apply subst_if; assumption.

(*while*)
elim (H e2 i t envl f H1).
intros e0' Hdef_e0'.
elim (H0 e2 i t envl f H1).
intros e1' Hdef_e1'.
exists (While e0' e1').
apply subst_while; assumption.

(*ref*)
elim (H e1 i t envl f H0).
intros e0' Hdef_e0'.
exists (Ref e0').
apply subst_ref; assumption.

(*deref*)
elim (H e1 i t envl f H0).
intros e0' Hdef_e0'.
exists (Deref e0').
apply subst_deref; assumption.

(*sequ*)
elim (H e2 i t envl f H1).
intros e0' Hdef_e0'.
elim (H0 e2 i t envl f H1).
intros e1' Hdef_e1'.
exists (Sequ e0' e1').
apply subst_sequ; assumption.

(*assign*)
elim (H e2 i t envl f H1).
intros e0' Hdef_e0'.
elim (H0 e2 i t envl f H1).
intros e1' Hdef_e1'.
exists (Assign e0' e1').
apply subst_assign; assumption.
Qed.

(*--------------------------------------*)

Parameter gen_loc : forall m : memory, exists l : location, ~ in_dom l m.

(*--------------------------------------*)
Lemma value_of_type_ref_is_loc :
 forall (e : expr) (t : type) (env : env_typ),
 is_value e -> type_of env e (Ref_t t) -> exists l : location, e = Loc l.
intros; inversion H.
rewrite <- H1 in H0; inversion H0.

exists l; trivial.

rewrite <- H1 in H0; inversion H0.
Qed.



(*--------------------------------------*)
Definition is_reducible (s : configuration) :=
  exists s' : configuration, red s s'.

(*--------------------------------------*)

Lemma memory_respects_equal_dom :
 forall (l : location) (dom : list location) (f' : identifier -> type)
   (f : location -> type) (m : memory),
 in_prop l dom ->
 memory_respects_env (empty_table f', intro_table dom f) m ->
 exists fm : location -> value, m = intro_table dom fm.

intros l dom f' f m.
elim m.
intros domm fm Hdom Hrespects.
exists fm.
inversion Hrespects.
rewrite H0.
trivial.
Qed.



(*--------------------------------------------*)
Lemma progression :
 forall (f : identifier -> type) (envl : env_typ_loc) 
   (a : expr) (t : type) (m : memory),
 type_of_config (empty_table f, envl) (a, m) t ->
 ~ is_value a -> is_reducible (a, m).

simple induction a; intros.
(*const*)
absurd (is_value (Const c)).
assumption.

apply Cst_val.

(*var*)
inversion H.
absurd (type_of (empty_table f, envl) (Var i) t).
intro abs.
inversion abs.
unfold empty_table in H7; inversion H7.
inversion H12.

assumption.

(*loc*)
absurd (is_value (Loc l)).
assumption.

apply Loc_val.

(*fun*)
absurd (is_value (Fun i e)).
assumption.

apply Fun_val.

(*apply*)
elim (is_value_dec e); intros He.
elim (is_value_dec e0); intros He0.
inversion He.

(*apply const*)
rewrite <- H3 in H1; inversion H1.
inversion H4.
inversion H9.
(*apply loc*)
rewrite <- H3 in H1; inversion H1.
inversion H4.
inversion H9.
(*apply fun *)
rewrite <- H3 in H1; inversion H1.
inversion H4.
elim (substitute_succeeds e1 e0 i t0 envl f H11).
intros e1' def_e1'.
unfold is_reducible in |- *.
exists (e1', m).
apply tete.
apply beta1.
assumption.

assumption.

(*apply e e0   o\u00f9 eO se reduit*)
inversion H1.
inversion H3.
elim (H0 t0 m (conj H10 H4) He0).
intros (e0', m') Hm.
unfold is_reducible in |- *.
exists (Apply e e0', m').

set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Apply e (id exp)) in |- *. 
replace (Apply e e0) with (ctx e0).
replace (Apply e e0') with (ctx e0').
replace (ctx e0, m) with (app_ctx ctx (e0, m)).
replace (ctx e0', m') with (app_ctx ctx (e0', m')).
apply cont.
assumption.

unfold ctx in |- *; apply app_right.
unfold id in |- *; apply hole.

assumption.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*apply e e0  o\u00f9  e se reduit*)
inversion H1.
inversion H3.
elim (H (Arrow t0 t) m (conj H8 H4) He).
intros (e'', m') Hm.
exists (Apply e'' e0, m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Apply (id exp) e0) in |- *. 
replace (Apply e e0) with (ctx e).
replace (Apply e'' e0) with (ctx e'').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e'', m') with (app_ctx ctx (e'', m')).
apply cont.
assumption.

unfold ctx in |- *; apply app_left.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*let in*)
inversion H1.
inversion H3.
elim (is_value_dec e); intros He.

(*e is value*)
elim (substitute_succeeds e0 e i t0 envl f).
intros e0' Hdef_e0'.
exists (e0', m).
apply tete.
apply beta2; assumption.
assumption.

(*e is not value*)
elim (H t0 m (conj H11 H4) He).
intros (er, mr) Hm'.
exists (Let_in i er e0, mr).
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Let_in i (id exp) e0) in |- *. 
replace (Let_in i e e0) with (ctx e).
replace (Let_in i er e0) with (ctx er).
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx er, mr) with (app_ctx ctx (er, mr)).
apply cont.
assumption.

unfold ctx in |- *; apply let_left.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*if*)
inversion H2.
inversion H4.
elim (is_value_dec e); intros He.
(*e is value*)
rewrite <- H9 in H10.
elim (bool_dec e env H10 He); intro He_val.
exists (e0, m).
apply tete.
rewrite He_val; apply red_if_true.

exists (e1, m).
apply tete.
rewrite He_val; apply red_if_false.

(*e is not value*)
elim (H (Const_t Bool) m (conj H10 H5) He).
intros (e', m') Hred_e_m.
exists (If e' e0 e1, m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => If (id exp) e0 e1) in |- *. 
replace (If e e0 e1) with (ctx e).
replace (If e' e0 e1) with (ctx e').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e', m') with (app_ctx ctx (e', m')).
apply cont.
assumption.

unfold ctx in |- *; apply if_left.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*while*)
exists (If e (Sequ e0 (While e e0)) (Const unit), m).
apply tete.
apply red_while.


(*ref*)
elim (is_value_dec e); intros He.
(*e is value*)
elim (gen_loc m).
intros l Hnot_l_in_m.
exists (Loc l, add_table l (value_intro e He) m).
apply tete.
apply red_ref.
assumption.

(*e is not value*)
inversion H0.
inversion H2.
elim (H t0 m (conj H6 H3) He).
intros (e', m') Hdef_e'.
exists (Ref e', m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Ref (id exp)) in |- *. 
replace (Ref e) with (ctx e).
replace (Ref e') with (ctx e').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e', m') with (app_ctx ctx (e', m')).
apply cont.
assumption.

unfold ctx in |- *; apply ref_ctx.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*deref*)
inversion H0.
elim (is_value_dec e); intros He.
(*e is value*)
inversion H2.
elim (value_of_type_ref_is_loc e t (empty_table f, envl) He H6).
intros l Hl.
rewrite Hl in H6; inversion H6.
inversion H9.
rewrite <- H14 in H3.

elim (memory_respects_equal_dom l dom f f0 m H13 H3).
intros fm Hm.

exists (value2expr (fm l), m).

apply tete.
rewrite Hm.
rewrite Hl.
apply red_deref.
simpl in |- *.
assumption.

apply intro_apply_table.
assumption.


(*e is not value*)
inversion H2.
elim (H (Ref_t t) m (conj H6 H3) He).
intros (e', m') Hred.
exists (Deref e', m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Deref (id exp)) in |- *. 
replace (Deref e) with (ctx e).
replace (Deref e') with (ctx e').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e', m') with (app_ctx ctx (e', m')).
apply cont.
assumption.

unfold ctx in |- *; apply deref_ctx.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*sequ*)
inversion H1.
elim (is_value_dec e); intros He.
(*e is value*)
exists (e0, m).
apply tete.
apply red_sequ.
assumption.

(*e is not value*)
inversion H3.
elim (H t' m (conj H8 H4) He).
intros (e', m') Hred.
exists (Sequ e' e0, m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Sequ (id exp) e0) in |- *. 
replace (Sequ e e0) with (ctx e).
replace (Sequ e' e0) with (ctx e').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e', m') with (app_ctx ctx (e', m')).
apply cont.
assumption.

unfold ctx in |- *; apply sequ_left.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(*assign*)
inversion H1.
inversion H3.
elim (is_value_dec e); intros He.
elim (is_value_dec e0); intros He0.

(*e and e0 value*)
rewrite <- H7 in H8. 
elim (value_of_type_ref_is_loc e t0 env He H8).
intros l Hl.
exists (Const unit, add_table l (value_intro e0 He0) m).
apply tete.
rewrite Hl; apply red_assign.
rewrite Hl in H8; inversion H8.
rewrite H7 in H12; inversion H12.
rewrite H17 in H14; inversion H14.
rewrite <- H18 in H4.
elim (memory_respects_equal_dom l dom f f0 m H15 H4).
intros fm Hm.
rewrite Hm; simpl in |- *.
assumption.

(* e0 not value*)
elim (H0 t0 m (conj H10 H4) He0).
intros (e0', m') He0_red_e0'.
rewrite <- H7 in H8.
elim (value_of_type_ref_is_loc e t0 env He H8).
intros l Hl.
rewrite Hl.
exists (Assign (Loc l) e0', m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Assign (Loc l) (id exp)) in |- *. 
replace (Assign (Loc l) e0) with (ctx e0).
replace (Assign (Loc l) e0') with (ctx e0').
replace (ctx e0, m) with (app_ctx ctx (e0, m)).
replace (ctx e0', m') with (app_ctx ctx (e0', m')).
apply cont.
assumption.

unfold ctx in |- *; apply assign_right.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

(* e not value*)
elim (H (Ref_t t0) m (conj H8 H4) He). 
intros (e', m') He_red_e'.
exists (Assign e' e0, m').
set (id := fun exp : expr => exp) in |- *.
set (ctx := fun exp : expr => Assign (id exp) e0) in |- *. 
replace (Assign e e0) with (ctx e).
replace (Assign e' e0) with (ctx e').
replace (ctx e, m) with (app_ctx ctx (e, m)).
replace (ctx e', m') with (app_ctx ctx (e', m')).
apply cont.
assumption.

unfold ctx in |- *; apply assign_left.
unfold id in |- *; apply hole.

simpl in |- *; trivial.
simpl in |- *; trivial.

unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.
unfold ctx in |- *; unfold id in |- *; simpl in |- *; trivial.

Qed.


(***********************************************************)
Theorem nf_typed_are_value :
 forall (f : identifier -> type) (envl : env_typ_loc) 
   (a : expr) (t : type) (m : memory),
 type_of_config (empty_table f, envl) (a, m) t ->
 ~ is_reducible (a, m) -> is_value a.
intros.
elim (is_value_dec a); intro.
assumption.

absurd (is_reducible (a, m)).
assumption.

apply progression with (t := t) (f := f) (envl := envl);
 (trivial; assumption; assumption).
Qed.



