(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec. 
Require Import Table.
Require Import Typing_spec.
Require Import Env_in_type_of_proof.


(*--------------- SUBSTITUTION LEMMA  ------------------*)
Lemma substitution_lemma :
 forall (e1 e2 e : expr) (i : identifier) (envx : env_typ_var)
   (envl : env_typ_loc) (t1 t : type),
 subst_expr e1 i e e2 ->
 type_of (add_table i t envx, envl) e1 t1 ->
 type_of (envx, envl) e t -> type_of (envx, envl) e2 t1.

intros e1.
induction e1
 as
  [c|
   i|
   l|
   i e1 Hrece1|
   e1_1 Hrece1_1 e1_0 Hrece1_0|
   i e1_1 Hrece1_1 e1_0 Hrece1_0|
   e1_1 Hrece1_1 e1_0 Hrece1_0 e1_2 Hrece1_2|
   e1_1 Hrece1_1 e1_0 Hrece1_0|
   e1 Hrece1|
   e1 Hrece1|
   e1_1 Hrece1_1 e1_0 Hrece1_0|
   e1_1 Hrece1_1 e1_0 Hrece1_0];
 intros e2 e x envx envl t1 t Hsubst Htypee1 Htypee. 

(* const *)
inversion Htypee1.
inversion Hsubst.
apply type_of_const.

(*var*)
inversion Hsubst.
rewrite <- H3.
replace t1 with t.
assumption.

apply add_table_correct with (t := envx) (a := x).
rewrite H1 in Htypee1; inversion Htypee1; assumption.

apply env_redundant with (x := x) (t := t).
intro Hfree.
inversion Hfree.
absurd (x = i); assumption.

assumption.

(* loc *)
inversion Hsubst.
inversion Htypee1.
apply type_of_loc.
assumption.

(* fun  *)
inversion Hsubst.
inversion Htypee1.
apply type_of_fun.
rewrite H2 in H9.
apply equiv_in_type_of with (envx' := add_table x t0 (add_table x t envx)).
apply equseless2; apply eqrefl.
assumption.

inversion Htypee1.
apply type_of_fun.
apply Hrece1 with (e := e) (i := x) (t := t).
assumption.
apply equiv_in_type_of with (envx' := add_table i t0 (add_table x t envx)).
apply eqswap.
assumption.
apply eqrefl.
assumption.
apply env_useless; assumption.


(*app*)
inversion Htypee1.
inversion Hsubst.
apply type_of_app with (t := t0).
apply Hrece1_1 with (e := e) (i := x) (t := t); assumption.

apply Hrece1_0 with (e := e) (i := x) (t := t); assumption.

(*let_in*)
inversion Htypee1.
inversion Hsubst.
(*x=i*)
apply type_of_let with (t := t0).
apply Hrece1_1 with (i := x) (e := e) (t := t); assumption.

rewrite H8 in H6.
apply equiv_in_type_of with (envx' := add_table x t0 (add_table x t envx)).
apply equseless2; apply eqrefl.
assumption.

(*~x=i*)
apply type_of_let with (t := t0).
apply Hrece1_1 with (i := x) (e := e) (t := t); assumption.

apply Hrece1_0 with (i := x) (e := e) (t := t).
assumption.

apply equiv_in_type_of with (envx' := add_table i t0 (add_table x t envx)).
apply eqswap.
assumption.
apply eqrefl.
assumption.

apply env_useless; assumption.

(*if*)
inversion Hsubst.
inversion Htypee1.
apply type_of_if.
apply Hrece1_1 with (i := x) (e := e) (t := t); assumption.

apply Hrece1_0 with (i := x) (e := e) (t := t); assumption.

apply Hrece1_2 with (i := x) (e := e) (t := t); assumption.



(*while*)
inversion Hsubst.
inversion Htypee1.
apply type_of_while with (t := t0).
apply Hrece1_1 with (e := e) (i := x) (t := t); assumption.

apply Hrece1_0 with (e := e) (i := x) (t := t); assumption.

(*ref*)
inversion Hsubst.
inversion Htypee1.
apply type_of_ref.
apply Hrece1 with (e := e) (i := x) (t := t); assumption.

(*deref*)
inversion Hsubst.
inversion Htypee1.
apply type_of_deref.
apply Hrece1 with (e := e) (i := x) (t := t); assumption.

(*sequ*)
inversion Hsubst.
inversion Htypee1.
apply type_of_sequ with (t' := t').
apply Hrece1_1 with (e := e) (i := x) (t := t); assumption.

apply Hrece1_0 with (e := e) (i := x) (t := t); assumption.

(*assign*)
inversion Hsubst.
inversion Htypee1.
apply type_of_assign with (t := t0).
apply Hrece1_1 with (e := e) (i := x) (t := t); assumption.

apply Hrece1_0 with (e := e) (i := x) (t := t); assumption.

Qed.